from .. import fixtures, config
from ..assertions import eq_

from sqlalchemy import Integer, String, select
from sqlalchemy import ForeignKey
from sqlalchemy import testing

from ..schema import Table, Column


class CTETest(fixtures.TablesTest):
    __backend__ = True
    __requires__ = 'ctes',

    run_inserts = 'each'
    run_deletes = 'each'

    @classmethod
    def define_tables(cls, metadata):
        Table("some_table", metadata,
              Column('id', Integer, primary_key=True),
              Column('data', String(50)),
              Column("parent_id", ForeignKey("some_table.id")))

        Table("some_other_table", metadata,
              Column('id', Integer, primary_key=True),
              Column('data', String(50)),
              Column("parent_id", Integer))

    @classmethod
    def insert_data(cls):
        config.db.execute(
            cls.tables.some_table.insert(),
            [
                {"id": 1, "data": "d1", "parent_id": None},
                {"id": 2, "data": "d2", "parent_id": 1},
                {"id": 3, "data": "d3", "parent_id": 1},
                {"id": 4, "data": "d4", "parent_id": 3},
                {"id": 5, "data": "d5", "parent_id": 3}
            ]
        )

    def test_select_nonrecursive_round_trip(self):
        some_table = self.tables.some_table

        with config.db.connect() as conn:
            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])).cte("some_cte")
            result = conn.execute(
                select([cte.c.data]).where(cte.c.data.in_(["d4", "d5"]))
            )
            eq_(result.fetchall(), [("d4", )])

    def test_select_recursive_round_trip(self):
        some_table = self.tables.some_table

        with config.db.connect() as conn:
            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])).cte(
                "some_cte", recursive=True)

            cte_alias = cte.alias("c1")
            st1 = some_table.alias()
            # note that SQL Server requires this to be UNION ALL,
            # can't be UNION
            cte = cte.union_all(
                select([st1]).where(st1.c.id == cte_alias.c.parent_id)
            )
            result = conn.execute(
                select([cte.c.data]).where(
                    cte.c.data != "d2").order_by(cte.c.data.desc())
            )
            eq_(
                result.fetchall(),
                [('d4',), ('d3',), ('d3',), ('d1',), ('d1',), ('d1',)]
            )

    def test_insert_from_select_round_trip(self):
        some_table = self.tables.some_table
        some_other_table = self.tables.some_other_table

        with config.db.connect() as conn:
            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])
            ).cte("some_cte")
            conn.execute(
                some_other_table.insert().from_select(
                    ["id", "data", "parent_id"],
                    select([cte])
                )
            )
            eq_(
                conn.execute(
                    select([some_other_table]).order_by(some_other_table.c.id)
                ).fetchall(),
                [(2, "d2", 1), (3, "d3", 1), (4, "d4", 3)]
            )

    @testing.requires.ctes_with_update_delete
    @testing.requires.update_from
    def test_update_from_round_trip(self):
        some_table = self.tables.some_table
        some_other_table = self.tables.some_other_table

        with config.db.connect() as conn:
            conn.execute(
                some_other_table.insert().from_select(
                    ['id', 'data', 'parent_id'],
                    select([some_table])
                )
            )

            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])
            ).cte("some_cte")
            conn.execute(
                some_other_table.update().values(parent_id=5).where(
                    some_other_table.c.data == cte.c.data
                )
            )
            eq_(
                conn.execute(
                    select([some_other_table]).order_by(some_other_table.c.id)
                ).fetchall(),
                [
                    (1, "d1", None), (2, "d2", 5),
                    (3, "d3", 5), (4, "d4", 5), (5, "d5", 3)
                ]
            )

    @testing.requires.ctes_with_update_delete
    @testing.requires.delete_from
    def test_delete_from_round_trip(self):
        some_table = self.tables.some_table
        some_other_table = self.tables.some_other_table

        with config.db.connect() as conn:
            conn.execute(
                some_other_table.insert().from_select(
                    ['id', 'data', 'parent_id'],
                    select([some_table])
                )
            )

            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])
            ).cte("some_cte")
            conn.execute(
                some_other_table.delete().where(
                    some_other_table.c.data == cte.c.data
                )
            )
            eq_(
                conn.execute(
                    select([some_other_table]).order_by(some_other_table.c.id)
                ).fetchall(),
                [
                    (1, "d1", None), (5, "d5", 3)
                ]
            )

    @testing.requires.ctes_with_update_delete
    def test_delete_scalar_subq_round_trip(self):

        some_table = self.tables.some_table
        some_other_table = self.tables.some_other_table

        with config.db.connect() as conn:
            conn.execute(
                some_other_table.insert().from_select(
                    ['id', 'data', 'parent_id'],
                    select([some_table])
                )
            )

            cte = select([some_table]).where(
                some_table.c.data.in_(["d2", "d3", "d4"])
            ).cte("some_cte")
            conn.execute(
                some_other_table.delete().where(
                    some_other_table.c.data ==
                    select([cte.c.data]).where(
                        cte.c.id == some_other_table.c.id)
                )
            )
            eq_(
                conn.execute(
                    select([some_other_table]).order_by(some_other_table.c.id)
                ).fetchall(),
                [
                    (1, "d1", None), (5, "d5", 3)
                ]
            )



from .. import fixtures, config
from ..assertions import eq_

from sqlalchemy import Integer, String
from ..schema import Table, Column


class SimpleUpdateDeleteTest(fixtures.TablesTest):
    run_deletes = 'each'
    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table('plain_pk', metadata,
              Column('id', Integer, primary_key=True),
              Column('data', String(50))
              )

    @classmethod
    def insert_data(cls):
        config.db.execute(
            cls.tables.plain_pk.insert(),
            [
                {"id": 1, "data": "d1"},
                {"id": 2, "data": "d2"},
                {"id": 3, "data": "d3"},
            ]
        )

    def test_update(self):
        t = self.tables.plain_pk
        r = config.db.execute(
            t.update().where(t.c.id == 2),
            data="d2_new"
        )
        assert not r.is_insert
        assert not r.returns_rows

        eq_(
            config.db.execute(t.select().order_by(t.c.id)).fetchall(),
            [
                (1, "d1"),
                (2, "d2_new"),
                (3, "d3")
            ]
        )

    def test_delete(self):
        t = self.tables.plain_pk
        r = config.db.execute(
            t.delete().where(t.c.id == 2)
        )
        assert not r.is_insert
        assert not r.returns_rows
        eq_(
            config.db.execute(t.select().order_by(t.c.id)).fetchall(),
            [
                (1, "d1"),
                (3, "d3")
            ]
        )

__all__ = ('SimpleUpdateDeleteTest', )



# mssql/information_schema.py
# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

# TODO: should be using the sys. catalog with SQL Server, not information
# schema

from ... import Table, MetaData, Column
from ...types import String, Unicode, UnicodeText, Integer, TypeDecorator
from ... import cast
from ... import util
from ...sql import expression
from ...ext.compiler import compiles

ischema = MetaData()


class CoerceUnicode(TypeDecorator):
    impl = Unicode

    def process_bind_param(self, value, dialect):
        if util.py2k and isinstance(value, util.binary_type):
            value = value.decode(dialect.encoding)
        return value

    def bind_expression(self, bindvalue):
        return _cast_on_2005(bindvalue)


class _cast_on_2005(expression.ColumnElement):
    def __init__(self, bindvalue):
        self.bindvalue = bindvalue


@compiles(_cast_on_2005)
def _compile(element, compiler, **kw):
    from . import base
    if compiler.dialect.server_version_info is None or \
             compiler.dialect.server_version_info < base.MS_2005_VERSION:
        return compiler.process(element.bindvalue, **kw)
    else:
        return compiler.process(cast(element.bindvalue, Unicode), **kw)

schemata = Table("SCHEMATA", ischema,
                 Column("CATALOG_NAME", CoerceUnicode, key="catalog_name"),
                 Column("SCHEMA_NAME", CoerceUnicode, key="schema_name"),
                 Column("SCHEMA_OWNER", CoerceUnicode, key="schema_owner"),
                 schema="INFORMATION_SCHEMA")

tables = Table("TABLES", ischema,
               Column("TABLE_CATALOG", CoerceUnicode, key="table_catalog"),
               Column("TABLE_SCHEMA", CoerceUnicode, key="table_schema"),
               Column("TABLE_NAME", CoerceUnicode, key="table_name"),
               Column(
                   "TABLE_TYPE", String(convert_unicode=True),
                   key="table_type"),
               schema="INFORMATION_SCHEMA")

columns = Table("COLUMNS", ischema,
                Column("TABLE_SCHEMA", CoerceUnicode, key="table_schema"),
                Column("TABLE_NAME", CoerceUnicode, key="table_name"),
                Column("COLUMN_NAME", CoerceUnicode, key="column_name"),
                Column("IS_NULLABLE", Integer, key="is_nullable"),
                Column("DATA_TYPE", String, key="data_type"),
                Column("ORDINAL_POSITION", Integer, key="ordinal_position"),
                Column("CHARACTER_MAXIMUM_LENGTH", Integer,
                       key="character_maximum_length"),
                Column("NUMERIC_PRECISION", Integer, key="numeric_precision"),
                Column("NUMERIC_SCALE", Integer, key="numeric_scale"),
                Column("COLUMN_DEFAULT", Integer, key="column_default"),
                Column("COLLATION_NAME", String, key="collation_name"),
                schema="INFORMATION_SCHEMA")

constraints = Table("TABLE_CONSTRAINTS", ischema,
                    Column("TABLE_SCHEMA", CoerceUnicode, key="table_schema"),
                    Column("TABLE_NAME", CoerceUnicode, key="table_name"),
                    Column("CONSTRAINT_NAME", CoerceUnicode,
                           key="constraint_name"),
                    Column("CONSTRAINT_TYPE", String(
                        convert_unicode=True), key="constraint_type"),
                    schema="INFORMATION_SCHEMA")

column_constraints = Table("CONSTRAINT_COLUMN_USAGE", ischema,
                           Column("TABLE_SCHEMA", CoerceUnicode,
                                  key="table_schema"),
                           Column("TABLE_NAME", CoerceUnicode,
                                  key="table_name"),
                           Column("COLUMN_NAME", CoerceUnicode,
                                  key="column_name"),
                           Column("CONSTRAINT_NAME", CoerceUnicode,
                                  key="constraint_name"),
                           schema="INFORMATION_SCHEMA")

key_constraints = Table("KEY_COLUMN_USAGE", ischema,
                        Column("TABLE_SCHEMA", CoerceUnicode,
                               key="table_schema"),
                        Column("TABLE_NAME", CoerceUnicode,
                               key="table_name"),
                        Column("COLUMN_NAME", CoerceUnicode,
                               key="column_name"),
                        Column("CONSTRAINT_NAME", CoerceUnicode,
                               key="constraint_name"),
                        Column("CONSTRAINT_SCHEMA", CoerceUnicode,
                               key="constraint_schema"),
                        Column("ORDINAL_POSITION", Integer,
                               key="ordinal_position"),
                        schema="INFORMATION_SCHEMA")

ref_constraints = Table("REFERENTIAL_CONSTRAINTS", ischema,
                        Column("CONSTRAINT_CATALOG", CoerceUnicode,
                               key="constraint_catalog"),
                        Column("CONSTRAINT_SCHEMA", CoerceUnicode,
                               key="constraint_schema"),
                        Column("CONSTRAINT_NAME", CoerceUnicode,
                               key="constraint_name"),
                        # TODO: is CATLOG misspelled ?
                        Column("UNIQUE_CONSTRAINT_CATLOG", CoerceUnicode,
                               key="unique_constraint_catalog"),

                        Column("UNIQUE_CONSTRAINT_SCHEMA", CoerceUnicode,
                               key="unique_constraint_schema"),
                        Column("UNIQUE_CONSTRAINT_NAME", CoerceUnicode,
                               key="unique_constraint_name"),
                        Column("MATCH_OPTION", String, key="match_option"),
                        Column("UPDATE_RULE", String, key="update_rule"),
                        Column("DELETE_RULE", String, key="delete_rule"),
                        schema="INFORMATION_SCHEMA")

views = Table("VIEWS", ischema,
              Column("TABLE_CATALOG", CoerceUnicode, key="table_catalog"),
              Column("TABLE_SCHEMA", CoerceUnicode, key="table_schema"),
              Column("TABLE_NAME", CoerceUnicode, key="table_name"),
              Column("VIEW_DEFINITION", CoerceUnicode, key="view_definition"),
              Column("CHECK_OPTION", String, key="check_option"),
              Column("IS_UPDATABLE", String, key="is_updatable"),
              schema="INFORMATION_SCHEMA")



from .. import fixtures, config
from ..config import requirements
from sqlalchemy import exc
from sqlalchemy import Integer, String, select, literal_column
from .. import assert_raises
from ..schema import Table, Column
from .. import provide_metadata
from .. import eq_


class ExceptionTest(fixtures.TablesTest):
    """Test basic exception wrapping.

    DBAPIs vary a lot in exception behavior so to actually anticipate
    specific exceptions from real round trips, we need to be conservative.

    """
    run_deletes = 'each'

    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table('manual_pk', metadata,
              Column('id', Integer, primary_key=True, autoincrement=False),
              Column('data', String(50))
              )

    @requirements.duplicate_key_raises_integrity_error
    def test_integrity_error(self):

        with config.db.connect() as conn:

            trans = conn.begin()
            conn.execute(
                self.tables.manual_pk.insert(),
                {'id': 1, 'data': 'd1'}
            )

            assert_raises(
                exc.IntegrityError,
                conn.execute,
                self.tables.manual_pk.insert(),
                {'id': 1, 'data': 'd1'}
            )

            trans.rollback()


class AutocommitTest(fixtures.TablesTest):

    run_deletes = 'each'

    __requires__ = 'autocommit',

    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table('some_table', metadata,
              Column('id', Integer, primary_key=True, autoincrement=False),
              Column('data', String(50)),
              test_needs_acid=True
              )

    def _test_conn_autocommits(self, conn, autocommit):
        trans = conn.begin()
        conn.execute(
            self.tables.some_table.insert(),
            {"id": 1, "data": "some data"}
        )
        trans.rollback()

        eq_(
            conn.scalar(select([self.tables.some_table.c.id])),
            1 if autocommit else None
        )

        conn.execute(self.tables.some_table.delete())

    def test_autocommit_on(self):
        conn = config.db.connect()
        c2 = conn.execution_options(isolation_level='AUTOCOMMIT')
        self._test_conn_autocommits(c2, True)
        conn.invalidate()
        self._test_conn_autocommits(conn, False)

    def test_autocommit_off(self):
        conn = config.db.connect()
        self._test_conn_autocommits(conn, False)


class EscapingTest(fixtures.TestBase):
    @provide_metadata
    def test_percent_sign_round_trip(self):
        """test that the DBAPI accommodates for escaped / nonescaped
        percent signs in a way that matches the compiler

        """
        m = self.metadata
        t = Table('t', m, Column('data', String(50)))
        t.create(config.db)
        with config.db.begin() as conn:
            conn.execute(t.insert(), dict(data="some % value"))
            conn.execute(t.insert(), dict(data="some %% other value"))

            eq_(
                conn.scalar(
                    select([t.c.data]).where(
                        t.c.data == literal_column("'some % value'"))
                ),
                "some % value"
            )

            eq_(
                conn.scalar(
                    select([t.c.data]).where(
                        t.c.data == literal_column("'some %% other value'"))
                ), "some %% other value"
            )
